from __future__ import print_function
from future.builtins import super
from lago import cmd
from lago.config import config as lago_config
from lago.lago_ansible import LagoAnsible
from lago import workdir as lago_workdir
from lago.log_utils import get_default_log_formatter
from sdk_utils import SDKWrapper, setup_sdk_logging
import weakref
import os
import logging


def add_stream_logger(level=logging.DEBUG, name=None):
    """
    Add a stream logger. This can be used for printing all SDK calls to stdout
    while working in an interactive session. Note this is a logger for the
    entire module, which will apply to all environments started in the same
    session. If you need a specific logger pass a ``logfile`` to
    :func:`~sdk.init`

    Args:
        level(int): :mod:`logging` log level
        name(str): logger name, will default to the root logger.

    Returns:
        None
    """

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(get_default_log_formatter())
    handler.setLevel(level)
    logger.addHandler(handler)


def init(config, workdir=None, logfile=None, loglevel=logging.INFO, **kwargs):
    """
    Initialize the Lago environment

    Args:
        config(str): Path to LagoInitFile
        workdir(str): Path to initalize the workdir, defaults to "$PWD/.lago"
        **kwargs(dict): Pass arguments to :func:`~lago.cmd.do_init`
        logfile(str): A path to setup a log file.
        loglevel(int): :mod:`logging` log level.

    Returns:
        :class:`~lago.sdk.SDK`: Initialized Lago enviornment

    Raises:
       :exc:`~lago.utils.LagoException`: If initialization failed
    """

    setup_sdk_logging(logfile, loglevel)
    defaults = lago_config.get_section('init')
    if workdir is None:
        workdir = os.path.abspath('.lago')
    defaults['workdir'] = workdir
    defaults['virt_config'] = config
    defaults.update(kwargs)
    workdir, prefix = cmd.do_init(**defaults)
    return SDK(workdir, prefix)


def load_env(workdir, logfile=None, loglevel=logging.INFO):
    """
    Load an existing Lago environment

    Args:
        workdir(str): Path to the workdir directory, as created by
        :func:`~lago.sdk.init` or created by the CLI.
        logfile(str): A Path to setup a log file.
        loglevel(int): :mod:`logging` log level.

    Returns:
        :class:`~lago.sdk.SDK`: Initialized Lago environment

    Raises:
       :exc:`~lago.utils.LagoException`: If loading the environment failed.
    """

    setup_sdk_logging(logfile, loglevel)
    workdir = os.path.abspath(workdir)
    loaded_workdir = lago_workdir.Workdir(path=workdir)
    prefix = loaded_workdir.get_prefix('current')
    return SDK(loaded_workdir, prefix)


class SDK(object):
    """
    The SDK can be initialized in 3 ways:

        1. (Preferred) - by calling :func:`sdk.init`.
        2. By loading an existing workdir from the disk, with
            :func:`~load_env`.
        3. By passing already created workdir and prefix objects.
    """

    def __init__(self, workdir, prefix):
        """
        __init__

        Args:
            workdir(:class:`~lago.workdir.Workdir`): The enviornment
                workdir.
            prefix(:class:~lago.prefix.Prefix): The enviornment Prefix.

        Returns:
            None
        """

        self._workdir = workdir
        self._prefix = prefix
        # Proxy object to the prefix to expose
        self._pprefix = SDKWrapper(weakref.proxy(self._prefix))

    def __getattr__(self, name):
        try:
            attr = super().__getattr__(name)
        except AttributeError:
            attr = getattr(self._pprefix, name)
        return attr

    def __dir__(self):
        # For auto-complete to work, we need to return the methods of the proxy
        # object as well, which we return in __getattr__.
        # As we cannot call self.__dir__, we need to construct 'self's
        # attributes.
        return sorted(
            set(dir(type(self)) + list(self.__dict__) + dir(self._pprefix))
        )

    def destroy(self):
        """
        Destroy the environment, this will terminate all resources, and remove
        entirely the Lago working directory.
        """
        self._workdir.destroy()

    def ansible_inventory_temp_file(
        self, keys=['vm-type', 'groups', 'vm-provider']
    ):
        """
        Context manager which returns Ansible inventory written on a tempfile.
        This is the same as :func:`~ansible_inventory`, only the inventory file
        is written to a tempfile.

        Args:
            keys (list of str): Path to the keys that will be used to
                create groups.

        Yields:
            tempfile.NamedTemporaryFile: Temp file containing the inventory
        """
        lansible = LagoAnsible(self._prefix)
        return lansible.get_inventory_temp_file(keys=keys)

    def ansible_inventory(
        self,
        keys=['vm-type', 'groups', 'vm-provider'],
    ):
        """
        Get an Ansible inventory as a string, ``keys`` should be list on which
        to group the hosts by. You can use any key defined in LagoInitFile.

        Examples of possible `keys`:

            `keys=['disks/0/metadata/arch']`, would group the hosts by the
            architecture.

            `keys=['/disks/0/metadata/distro', 'disks/0/metadata/arch']`,
                would create groups by architecture and also by distro.

            `keys=['groups']` - would group hosts by the groups defined for
                each VM in the LagoInitFile, i.e.:

                    domains:

                        vm-01:
                            ...
                            groups: web-server
                            ..
                        vm-02:
                            ..
                            groups: db-server


        Args:
            keys (list of str): Path to the keys that will be used to
                create groups.

        Returns:
            str: INI-like Ansible inventory
        """

        lansible = LagoAnsible(self._prefix)
        return lansible.get_inventory_str(keys=keys)



#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import logging
from textwrap import dedent
from lago.validation import check_import

LOGGER = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.WARNING)
LOGGER.addHandler(ch)

if not check_import('guestfs'):
    pip_link = 'http://libguestfs.org/download/python/guestfs-1.XX.YY.tar.gz'

    msg = dedent(
        """
            WARNING: - guestfs not found. Some Lago features will not work.
            Please install it either by using your distribution package, or
            with pip.

            For Fedora/CentOS, run: yum install python2-libguestfs

            For Debian, run: apt-get install python-libguestfs

            For pip, go to http://libguestfs.org/download/python, pick
            the version and run:

            pip install {pip_link}
        """.format(pip_link=pip_link)
    )
    LOGGER.warning(msg)



#
# Copyright 2014 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
"""
VM Plugins
============
There are two VM-related plugin extension points, there's the
VM Type Plugin, that allows you to modify at a higher level the inner
workings of the VM class (domain concept in the initfile).
The other plugin extension point, the [VM Provider Plugin], that allows you to
create an alternative implementation of the provisioning details for the VM,
for example, using a remote libvirt instance or similar.
"""
from copy import deepcopy
import contextlib
import functools
import logging
import os
import warnings
from abc import (ABCMeta, abstractmethod)

from scp import SCPClient, SCPException

from .. import (
    utils,
    log_utils,
    plugins,
    ssh,
)
from lago.config import config

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)


class VMError(Exception):
    pass


class ExtractPathError(VMError):
    pass


class ExtractPathNoPathError(VMError):
    pass


def check_defined(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.defined():
            raise RuntimeError('VM %s is not defined' % self.vm.name())
        return func(self, *args, **kwargs)

    return wrapper


def check_alive(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.alive():
            raise RuntimeError('VM %s is not running' % self.name())
        return func(self, *args, **kwargs)

    return wrapper


class VMProviderPlugin(plugins.Plugin):
    """
    If you want to use a custom provider for you VMs (say, ovirt for example),
    you have to inherit from this class, and then define the
    'default_vm_provider' in your config to be your plugin, or explicitly
    specify it on each domain definition in the initfile with 'vm-provider' key

    You will have to override at least all the abstractmethods in order to
    write a provider plugin, even if they are just runnig `pass`.
    """

    def __init__(self, vm):
        self.vm = vm

    @abstractmethod
    def start(self, *args, **kwargs):
        """
        Start a domain

        Returns:
            None
        """
        pass

    @abstractmethod
    def stop(self, *args, **kwargs):
        """
        Stop a domain

        Returns:
            None
        """
        pass

    @abstractmethod
    def shutdown(self, *args, **kwargs):
        """
        Shutdown a domain

        Returns:
            None
        """
        pass

    @abstractmethod
    def reboot(self, *args, **kwargs):
        """
        Reboot a domain

        Returns:
            None
        """
        pass

    @abstractmethod
    def defined(self, *args, **kwargs):
        """
        Return if the domain is defined (libvirt concept), currently used only
        by the libvirt provider, put here to allow backwards compatibility.

        Returns:
            bool: True if the domain is already defined (libvirt concept)
        """
        pass

    @abstractmethod
    def bootstrap(self, *args, **kwargs):
        """
        Does any actions needed to get the domain ready to be used, ran on
        prefix init.

        Return:
            None
        """
        pass

    @abstractmethod
    def state(self, *args, **kwargs):
        """
        Return the current state of the domain

        Returns:
            str: Small description of the current domain state
        """
        pass

    @abstractmethod
    def create_snapshot(self, name, *args, **kwargs):
        """
        Take any actions needed to create a snapshot

        Args:
            name(str): Name for the snapshot, will be used as key to retrieve
                it later

        Returns:
            None
        """
        pass

    @abstractmethod
    def revert_snapshot(self, name, *args, **kwargs):
        """
        Take any actions needed to revert/restore a snapshot

        Args:
            name(str): Name for the snapshot, same that was set on creation

        Returns:
            None
        """
        pass

    def export_disks(self, standalone, dst_dir, compress, *args, **kwargs):
        """
        Export 'disks' as a standalone image or a layered image.

        Args:
            disks(list): The names of the disks to export
              (None means all the disks)
            standalone(bool): If true create a copy of the layered image
              else create a new disk which is a combination of the current
              layer and the base disk.
            dst_dir (str): dir to place the exported images
            compress (bool): if true, compress the exported image.
        """
        pass

    def interactive_console(self):
        """
        Run an interactive console

        Returns:
            lago.utils.CommandStatus: resulf of the interactive execution
        """
        return self.vm.interactive_ssh()

    def extract_paths_dead(self, paths, ignore_nopath):
        """
        Extract the given paths from the domain, without the underlying OS
        awareness
        """
        pass

    def extract_paths(self, paths, ignore_nopath):
        """
        Extract the given paths from the domain

        Args:
            paths(list of str): paths to extract
            ignore_nopath(boolean): if True will ignore none existing paths.

        Returns:
            None

        Raises:
            :exc:`~lago.plugins.vm.ExtractPathNoPathError`: if a none existing
                path was found on the VM, and ``ignore_nopath`` is True.
            :exc:`~lago.plugins.vm.ExtractPathError`: on all other failures.
        """
        if self.vm.alive() and self.vm.ssh_reachable(
            tries=5, propagate_fail=False
        ):
            self._extract_paths_scp(paths=paths, ignore_nopath=ignore_nopath)
        else:
            raise ExtractPathError(
                'Unable to extract paths from {0}: unreachable with SSH'.
                format(self.vm.name())
            )

    def _extract_paths_scp(self, paths, ignore_nopath):
        for host_path, guest_path in paths:
            LOGGER.debug(
                'Extracting scp://%s:%s to %s',
                self.vm.name(),
                host_path,
                guest_path,
            )
            try:
                self.vm.copy_from(
                    local_path=guest_path,
                    remote_path=host_path,
                    propagate_fail=False
                )
            except SCPException as err:
                err_substr = ': No such file or directory'
                if len(err.args) > 0 and isinstance(
                    err.args[0], basestring
                ) and err_substr in err.args[0]:
                    if ignore_nopath:
                        LOGGER.debug('%s: ignoring', err.args[0])
                    else:
                        raise ExtractPathNoPathError(err.args[0])
                else:
                    raise


class VMPlugin(plugins.Plugin):
    __metaclass__ = ABCMeta
    '''
    This class takes care of the high level abstraction for a VM (a domain in
    the initfile lingo). From starting/stopping it to loading and calling the
    provider if needed. If you want to change only the way the VM is
    provisioned you can take a look to the `class:VMProviderPlugin` instead.

    This base class includes also some basic methods implemented with ssh.

    VM properties:
    * name
    * cpus
    * memory
    * disks
    * metadata
    * network/mac addr
    * virt_env
    '''

    def __init__(self, env, spec):
        self.virt_env = env
        self._spec = self._normalize_spec(spec.copy())

        self._ssh_client = None
        self.service_providers = plugins.load_plugins(
            namespace=plugins.PLUGIN_ENTRY_POINTS['vm-service'],
            instantiate=False,
        )
        self._service_class = self._get_service_provider()
        self.vm_providers = plugins.load_plugins(
            namespace=plugins.PLUGIN_ENTRY_POINTS['vm-provider'],
            instantiate=False,
        )
        self.provider = self._get_vm_provider()

    def start(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.start(*args, **kwargs)

    def stop(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.stop(*args, **kwargs)

    def shutdown(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.shutdown(*args, **kwargs)

    def reboot(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.reboot(*args, **kwargs)

    def defined(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.defined(*args, **kwargs)

    def bootstrap(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.bootstrap(*args, **kwargs)

    def state(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.state(*args, **kwargs)

    def create_snapshot(self, name, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.create_snapshot(name, *args, **kwargs)

    def revert_snapshot(self, name, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.revert_snapshot(name, *args, **kwargs)

    def interactive_console(self, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.interactive_console(*args, **kwargs)

    def extract_paths(self, paths, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.extract_paths(paths, *args, **kwargs)

    def extract_paths_dead(self, paths, *args, **kwargs):
        """
        Thin method that just uses the provider
        """
        return self.provider.extract_paths_dead(paths, *args, **kwargs)

    def export_disks(
        self,
        standalone=True,
        dst_dir=None,
        compress=False,
        collect_only=False,
        with_threads=True,
        *args,
        **kwargs
    ):
        """
        Thin method that just uses the provider
        """
        return self.provider.export_disks(
            standalone, dst_dir, compress, collect_only, with_threads, *args,
            **kwargs
        )

    def copy_to(self, local_path, remote_path, recursive=True):
        with LogTask(
            'Copy %s to %s:%s' % (local_path, self.name(), remote_path),
        ):
            with self._scp() as scp:
                scp.put(
                    files=local_path,
                    remote_path=remote_path,
                    recursive=recursive,
                )

    def copy_from(
        self, remote_path, local_path, recursive=True, propagate_fail=True
    ):
        with self._scp(propagate_fail=propagate_fail) as scp:
            scp.get(
                recursive=recursive,
                remote_path=remote_path,
                local_path=local_path,
            )

    @property
    def metadata(self):
        return self._spec['metadata'].copy()

    @property
    def disks(self):
        return self._spec['disks'][:]

    @property
    def spec(self):
        return deepcopy(self._spec)

    @property
    def mgmt_name(self):
        return self._spec.get('mgmt_net', None)

    @property
    def mgmt_net(self):
        return self.virt_env.get_net(name=self.mgmt_name)

    @property
    def vm_type(self):
        return self._spec['vm-type']

    @property
    def groups(self):
        """
        Returns:
            list of str: The names of the groups to which this vm belongs
                (as specified in the init file)
        """
        groups = self._spec.get('groups', [])
        if groups:
            return groups[:]
        else:
            return groups

    def name(self):
        return str(self._spec['name'])

    def iscsi_name(self):
        return 'iqn.2014-07.org.lago:%s' % self.name()

    def ip(self):
        res = self.mgmt_net.resolve(self.name())
        return res.encode('ascii', 'ignore')

    def all_ips(self):
        nets = {}
        ips = []
        nets = self.virt_env.get_nets()
        for net in nets.values():
            mapping = net.mapping()
            for hostname, ip in mapping.items():
                # hostname is <hostname>-<ifacename>
                if hostname.startswith(self.name() + "-"):
                    ips.append(str(ip))
        return ips

    def ips_in_net(self, net_name):
        ips = []
        net = self.virt_env.get_net(name=net_name)
        mapping = net.mapping()
        for hostname, ip in mapping.items():
            # hostname is <hostname>-<ifacename>
            if hostname.startswith(self.name() + "-"):
                ips.append(str(ip))
        return ips

    def ssh(
        self,
        command,
        data=None,
        show_output=True,
        propagate_fail=True,
        tries=None,
    ):
        if not self.alive():
            raise RuntimeError('Attempt to ssh into a not running host')

        return ssh.ssh(
            ip_addr=self.ip(),
            host_name=self.name(),
            command=command,
            data=None,
            show_output=True,
            propagate_fail=True,
            tries=None,
            ssh_key=self.virt_env.prefix.paths.ssh_id_rsa(),
            username=self._spec.get('ssh-user'),
            password=self._spec.get('ssh-password'),
        )

    def wait_for_ssh(self):
        return ssh.wait_for_ssh(
            ip_addr=self.ip(),
            host_name=self.name(),
            connect_timeout=self._spec.get('boot_time_sec', 600),
            ssh_key=self.virt_env.prefix.paths.ssh_id_rsa(),
            username=self._spec.get('ssh-user'),
            password=self._spec.get('ssh-password'),
        )

    def ssh_script(self, path, show_output=True):
        return ssh.ssh_script(
            ip_addr=self.ip(),
            host_name=self.name(),
            ssh_key=self.virt_env.prefix.paths.ssh_id_rsa(),
            path=path,
            show_output=show_output,
            username=self._spec.get('ssh-user'),
            password=self._spec.get('ssh-password'),
        )

    def alive(self):
        return self.state() == 'running'

    @check_alive
    def ssh_reachable(self, tries=None, propagate_fail=True):
        """
        Check if the VM is reachable with ssh

        Args:
            tries(int): Number of tries to try connecting to the host
            propagate_fail(bool): If set to true, this event will appear
            in the log and fail the outter stage. Otherwise, it will be
            discarded.

        Returns:
            bool: True if the VM is reachable.
        """

        try:
            ssh.get_ssh_client(
                ip_addr=self.ip(),
                host_name=self.name(),
                ssh_tries=tries,
                propagate_fail=propagate_fail,
                ssh_key=self.virt_env.prefix.paths.ssh_id_rsa(),
                username=self._spec.get('ssh-user'),
                password=self._spec.get('ssh-password'),
            )
        except RuntimeError:
            return False

        return True

    def save(self, path=None):
        if path is None:
            path = self.virt_env.virt_path('vm-%s' % self.name())

        dst_dir = os.path.dirname(path)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)

        with open(path, 'w') as f:
            utils.json_dump(self._spec, f)

    @check_alive
    def service(self, name):
        if self._service_class is None:
            self._detect_service_provider()

        return self._service_class(self, name)

    @check_alive
    def interactive_ssh(self, command=None):
        if command is None:
            command = ['bash']

        return ssh.interactive_ssh(
            ip_addr=self.ip(),
            host_name=self.name(),
            ssh_key=self.virt_env.prefix.paths.ssh_id_rsa(),
            command=command,
            username=self._spec.get('ssh-user'),
            password=self._spec.get('ssh-password'),
        )

    def nics(self):
        return self._spec['nics'][:]

    def nets(self):
        return [nic['net'] for nic in self._spec['nics']]

    def distro(self):
        distro = self._spec.get('distro', None)
        if distro is None:
            distro = self._template_metadata().get('distro', None)

        return distro

    def root_password(self):
        root_password = self._spec.get('root-password', None)
        if root_password is None:
            root_password = self._spec.get('ssh-password', '')

        return root_password

    def collect_artifacts(self, host_path, ignore_nopath):
        self.extract_paths(
            [
                (
                    guest_path,
                    os.path.join(host_path, guest_path.replace('/', '_')),
                ) for guest_path in self._artifact_paths()
            ],
            ignore_nopath=ignore_nopath
        )

    def guest_agent(self):
        if 'guest-agent' not in self._spec:
            for possible_name in ('qemu-guest-agent', 'qemu-ga'):
                try:
                    if self.service(possible_name).exists():
                        self._spec['guest-agent'] = possible_name
                        self.save()
                        break
                except RuntimeError as err:
                    raise RuntimeError(
                        'Could not find guest agent service: %s' % err
                    )
            else:
                raise RuntimeError('Could not find guest agent service')

        return self.service(self._spec['guest-agent'])

    def has_guest_agent(self):
        try:
            self.guest_agent()
        except RuntimeError:
            return False

        return True

    def _get_vm_provider(self):
        default_provider = config.get('default_vm_provider')
        provider_name = self._spec.get('vm-provider', default_provider)
        provider = self.vm_providers.get(provider_name)
        self._spec['vm-provider'] = provider_name
        return provider(vm=self)

    @classmethod
    def _normalize_spec(cls, spec):
        spec['snapshots'] = spec.get('snapshots', {})
        spec['metadata'] = spec.get('metadata', {})

        if 'root-password' not in spec:
            root_password = config.get('root_password')
            if root_password:
                spec['ssh-password'] = root_password
            else:
                spec['ssh-password'] = config.get('ssh_password')

        if 'ssh-user' not in spec:
            spec['ssh-user'] = config.get('ssh_user')

        return spec

    @contextlib.contextmanager
    def _scp(self, propagate_fail=True):
        client = ssh.get_ssh_client(
            propagate_fail=propagate_fail,
            ip_addr=self.ip(),
            host_name=self.name(),
            ssh_key=self.virt_env.prefix.paths.ssh_id_rsa(),
            username=self._spec.get('ssh-user'),
            password=self._spec.get('ssh-password'),
        )
        scp = SCPClient(client.get_transport())
        try:
            yield scp
        finally:
            client.close()

    def _detect_service_provider(self):
        LOGGER.debug('Detecting service provider for %s', self.name())

        for provider_name, service_class in self.service_providers.items():
            if service_class.is_supported(self):
                LOGGER.debug(
                    'Setting %s as service provider for %s',
                    provider_name,
                    self.name(),
                )
                self._service_class = service_class
                self._spec['service_provider'] = provider_name
                self.save()
                return

        raise RuntimeError('No service provider detected for %s' % self.name())

    def _template_metadata(self):
        return self._spec['disks'][0].get('metadata', {})

    def _artifact_paths(self):
        return self._spec.get('artifacts', [])

    def _get_service_provider(self):
        """
        **NOTE**: Can be reduced to just one get call once we remove support
        for the service_class spec entry

        Returns:
            class: class for the loaded provider for that vm_spec
            None: if no provider was specified in the vm_spec
        """
        service_class = self._spec.get('service_class', None)
        if service_class is not None:
            warnings.warn(
                'The service_class key for a domain is deprecated, you should '
                'change it to service_provider instead'
            )
            service_provider = _resolve_service_class(
                class_name=service_class,
                service_providers=self.service_providers,
            )
        else:
            service_provider = self.service_providers.get(
                self._spec.get('service_provider', None),
                None,
            )

        return service_provider


def _resolve_service_class(class_name, service_providers):
    """
    **NOTE**: This must be remved once the service_class spec entry is fully
    deprecated

    Retrieves a service plugin class from the class name instead of the
    provider name

    Args:
        class_name(str): Class name of the service plugin to retrieve
        service_providers(dict): provider_name->provider_class of the loaded
            service providers

    Returns:
        class: Class of the plugin that matches that name

    Raises:
        lago.plugins.NoSuchPluginError: if there was no service plugin that
            matched the search
    """
    for plugin in service_providers.itervalues():
        if plugin.__class__.__name__ == class_name:
            return plugin

    raise plugins.NoSuchPluginError(
        'No service provider plugin with class name %s found, loaded '
        'providers: %s' % (
            class_name,
            [
                plugin.__class__.__name__
                for plugin in service_providers.itervalues()
            ],
        )
    )



#
# Copyright 2014-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
from future.builtins import super
from collections import defaultdict
import functools
import logging
import time
from copy import deepcopy

from lxml import etree as ET
import lago.providers.libvirt.utils as libvirt_utils
from lago import brctl, log_utils, utils
from lago.config import config

LOGGER = logging.getLogger(__name__)
LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)


class Network(object):
    def __init__(self, env, spec, compat):
        self._env = env
        libvirt_url = config.get('libvirt_url')
        self.libvirt_con = libvirt_utils.get_libvirt_connection(
            name=env.uuid + libvirt_url,
            libvirt_url=libvirt_url,
        )
        self._spec = spec
        self.compat = compat

    def __del__(self):
        if self.libvirt_con is not None:
            self.libvirt_con.close()

    def name(self):
        return self._spec['name']

    def gw(self):
        return self._spec.get('gw')

    def mtu(self):
        if self.libvirt_con.getLibVersion() > 3001001:
            return self._spec.get('mtu', '1500')
        else:
            return '1500'

    def is_management(self):
        return self._spec.get('management', False)

    def add_mappings(self, mappings):
        for name, ip, mac in mappings:
            self.add_mapping(name, ip, save=False)
        self.save()

    def add_mapping(self, name, ip, save=True):
        self._spec['mapping'][name] = ip
        if save:
            self.save()

    def resolve(self, name):
        return self._spec['mapping'][name]

    def mapping(self):
        return self._spec['mapping']

    def _libvirt_name(self):
        return self._env.prefixed_name(self.name(), max_length=15)

    def _libvirt_xml(self):
        raise NotImplementedError(
            'should be implemented by the specific network class'
        )

    def alive(self):
        net_names = [net.name() for net in self.libvirt_con.listAllNetworks()]
        return self._libvirt_name() in net_names

    def start(self, attempts=5, timeout=2):
        """
        Start the network, will check if the network is active ``attempts``
        times, waiting ``timeout`` between each attempt.

        Args:
            attempts (int): number of attempts to check the network is active
            timeout  (int): timeout for each attempt

        Returns:
            None

        Raises:
            RuntimeError: if network creation failed, or failed to verify it is
            active.
        """

        if not self.alive():
            with LogTask('Create network %s' % self.name()):
                net = self.libvirt_con.networkCreateXML(self._libvirt_xml())
                if net is None:
                    raise RuntimeError(
                        'failed to create network, XML: %s' %
                        (self._libvirt_xml())
                    )
                for _ in range(attempts):
                    if net.isActive():
                        return
                    LOGGER.debug(
                        'waiting for network %s to become active', net.name()
                    )
                    time.sleep(timeout)
                raise RuntimeError(
                    'failed to verify network %s is active' % net.name()
                )

    def stop(self):
        if self.alive():
            with LogTask('Destroy network %s' % self.name()):
                self.libvirt_con.networkLookupByName(
                    self._libvirt_name(),
                ).destroy()

    def save(self):
        with open(self._env.virt_path('net-%s' % self.name()), 'w') as f:
            utils.json_dump(self._spec, f)

    @property
    def spec(self):
        return deepcopy(self._spec)


class NATNetwork(Network):
    def _generate_dns_forward(self, forward_ip):
        dns = ET.Element('dns', forwardPlainNames='yes')
        dns.append(ET.Element('forwarder', addr=forward_ip))
        return dns

    def _generate_dns_disable(self):
        dns = ET.Element('dns', enable='no')
        return dns

    def _generate_main_dns(self, records, subnet, forward_plain='no'):
        dns = ET.Element('dns', forwardPlainNames=forward_plain)
        reverse_records = defaultdict(list)
        ipv6_prefix = self._ipv6_prefix(subnet=subnet)
        for hostname, ip in records.iteritems():
            reverse_records[ip] = reverse_records[ip] + [hostname]
        for ip, hostnames in reverse_records.iteritems():
            record_ipv4 = ET.Element('host', ip=ip)
            record_ipv6 = ET.Element('host', ip=ipv6_prefix + ip)
            for hostname in sorted(hostnames):
                host = ET.Element('hostname')
                host.text = hostname
                record_ipv4.append(host)
                record_ipv6.append(deepcopy(host))
            dns.append(record_ipv4)
            dns.append(record_ipv6)

        return dns

    def _ipv6_prefix(self, subnet, const='fd8f:1391:3a82:'):
        return '{0}{1}::'.format(const, subnet)

    def _libvirt_xml(self):
        net_raw_xml = libvirt_utils.get_template('net_nat_template.xml')

        subnet = self.gw().split('.')[2]
        ipv6_prefix = self._ipv6_prefix(subnet=subnet)
        mtu = self.mtu()

        replacements = {
            '@NAME@': self._libvirt_name(),
            '@BR_NAME@': ('%s-nic' % self._libvirt_name())[:12],
            '@GW_ADDR@': self.gw(),
            '@SUBNET@': subnet
        }
        for k, v in replacements.items():
            net_raw_xml = net_raw_xml.replace(k, v, 1)

        parser = ET.XMLParser(remove_blank_text=True)
        net_xml = ET.fromstring(net_raw_xml, parser)

        if mtu != '1500':
            net_xml.append(ET.Element(
                'mtu',
                size=str(mtu),
            ))
        if 'dhcp' in self._spec:
            ipv4 = net_xml.xpath('/network/ip')[0]
            ipv6 = net_xml.xpath('/network/ip')[1]

            def make_ipv4(last):
                return '.'.join(self.gw().split('.')[:-1] + [str(last)])

            dhcp = ET.Element('dhcp')
            dhcpv6 = ET.Element('dhcp')
            ipv4.append(dhcp)
            ipv6.append(dhcpv6)

            dhcp.append(
                ET.Element(
                    'range',
                    start=make_ipv4(self._spec['dhcp']['start']),
                    end=make_ipv4(self._spec['dhcp']['end']),
                )
            )
            dhcpv6.append(
                ET.Element(
                    'range',
                    start=ipv6_prefix + make_ipv4(self._spec['dhcp']['start']),
                    end=ipv6_prefix + make_ipv4(self._spec['dhcp']['end']),
                )
            )

            for hostname, ip4 in self._spec['mapping'].items():
                dhcp.append(
                    ET.Element(
                        'host',
                        mac=utils.ipv4_to_mac(ip4),
                        ip=ip4,
                        name=hostname
                    )
                )
                dhcpv6.append(
                    ET.Element(
                        'host',
                        id='0:3:0:1:' + utils.ipv4_to_mac(ip4),
                        ip=ipv6_prefix + ip4,
                        name=hostname
                    )
                )

        if utils.ver_cmp(self.compat, '0.36.11') >= 0:
            if self.is_management():
                domain_xml = ET.Element(
                    'domain',
                    name=self._spec['dns_domain_name'],
                    localOnly='yes'
                )
                net_xml.append(domain_xml)
                net_xml.append(
                    self._generate_main_dns(self._spec['dns_records'], subnet)
                )
            else:
                if self.libvirt_con.getLibVersion() < 2002000:
                    net_xml.append(
                        self._generate_dns_forward(self._spec['dns_forward'])
                    )
                else:
                    net_xml.append(self._generate_dns_disable())
        else:
            LOGGER.debug(
                'Generating network XML with compatibility prior to %s',
                self.compat
            )
            # Prior to v0.37, DNS records were only the  mappings of the
            # management network.
            if self.is_management():
                if 'dns_domain_name' in self._spec:
                    domain_xml = ET.Element(
                        'domain',
                        name=self._spec['dns_domain_name'],
                        localOnly='yes'
                    )
                    net_xml.append(domain_xml)

                net_xml.append(
                    self._generate_main_dns(
                        self._spec['mapping'], subnet, forward_plain='yes'
                    )
                )
            else:
                dns = ET.Element('dns', forwardPlainNames='yes', enable='yes')
                net_xml.append(dns)

        LOGGER.debug(
            'Generated Network XML\n {0}'.format(
                ET.tostring(net_xml, pretty_print=True)
            )
        )
        return ET.tostring(net_xml)


class BridgeNetwork(Network):
    def _libvirt_xml(self):
        net_raw_xml = libvirt_utils.get_template('net_br_template.xml')

        replacements = {
            '@NAME@': self._libvirt_name(),
            '@BR_NAME@': self._libvirt_name(),
        }
        for k, v in replacements.items():
            net_raw_xml = net_raw_xml.replace(k, v, 1)

        return net_raw_xml

    def start(self):
        if brctl.exists(self._libvirt_name()):
            return

        brctl.create(self._libvirt_name())
        try:
            super().start()
        except:
            brctl.destroy(self._libvirt_name())

    def stop(self):
        super().stop()
        if brctl.exists(self._libvirt_name()):
            brctl.destroy(self._libvirt_name())



import os
import subprocess
from setuptools import setup


def check_output(args):
    proc = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = proc.communicate()

    if proc.returncode:
        raise RuntimeError(
            'Failed to run %s\nrc=%s\nstdout=\n%sstderr=%s' %
            (args, proc.returncode, stdout, stderr)
        )

    return stdout


def get_version(project_dir=os.curdir):
    """
    Retrieves the version of the package, from the PKG-INFO file or generates
    it with the version script
    Returns:
        str: Version for the package
    Raises:
        RuntimeError: If the version could not be retrieved
    """
    if 'LAGO_VERSION' in os.environ and os.environ['LAGO_VERSION']:
        return os.environ['LAGO_VERSION']

    version = None
    pkg_info_file = os.path.join(project_dir, 'PKG-INFO')
    version_manager = os.path.join(project_dir, 'scripts/version_manager.py')
    if os.path.exists(pkg_info_file):
        with open(pkg_info_file) as info_fd:
            for line in info_fd.readlines():
                if line.startswith('Version: '):
                    version = line.split(' ', 1)[-1]

    elif os.path.exists(version_manager):
        version = check_output([version_manager, project_dir,
                                'version']).strip()

    if version is None:
        raise RuntimeError('Failed to get package version')

    # py3 compatibility step
    if not isinstance(version, str) and isinstance(version, bytes):
        version = version.decode()

    return version


if __name__ == '__main__':
    os.environ['PBR_VERSION'] = get_version()
    setup(
        setup_requires=['pbr', 'dulwich'],
        pbr=True,
    )



#
# Copyright 2016 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

from future.builtins import super
from lago.plugins.service import (
    ServicePlugin,
    ServiceState,
)


class SystemdService(ServicePlugin):
    BIN_PATH = '/usr/bin/systemctl'

    def _request_start(self):
        super()._request_start()
        return self._vm.ssh([self.BIN_PATH, 'start', self._name])

    def _request_stop(self):
        super()._request_stop()
        return self._vm.ssh([self.BIN_PATH, 'stop', self._name])

    def state(self):
        ret = self._vm.ssh([self.BIN_PATH, 'status --lines=0', self._name])
        if not ret:
            return ServiceState.ACTIVE

        lines = [l.strip() for l in ret.out.split('\n')]
        loaded = [l for l in lines if l.startswith('Loaded:')].pop()

        if loaded.split()[1] == 'loaded':
            return ServiceState.INACTIVE

        return ServiceState.MISSING


class SysVInitService(ServicePlugin):
    BIN_PATH = '/sbin/service'

    def _request_start(self):
        super()._request_start()
        return self._vm.ssh([self.BIN_PATH, self._name, 'start'])

    def _request_stop(self):
        super()._request_stop()
        return self._vm.ssh([self.BIN_PATH, self._name, 'stop'])

    def state(self):
        ret = self._vm.ssh([self.BIN_PATH, self._name, 'status'])

        if ret.code == 0:
            return ServiceState.ACTIVE

        if ret.out.strip().endswith('is stopped'):
            return ServiceState.INACTIVE

        return ServiceState.MISSING


class SystemdContainerService(ServicePlugin):
    BIN_PATH = '/usr/bin/docker'
    HOST_BIN_PATH = '/usr/bin/systemctl'

    def _request_start(self):
        super()._request_start()
        ret = self._vm.ssh(
            [self.BIN_PATH, 'exec vdsmc systemctl start', self._name]
        )

        if ret.code == 0:
            return ret

        return self._vm.ssh([self.HOST_BIN_PATH, 'start', self._name])

    def _request_stop(self):
        super()._request_stop()
        ret = self._vm.ssh(
            [self.BIN_PATH, 'exec vdsmc systemctl stop', self._name]
        )

        if ret.code == 0:
            return ret

        return self._vm.ssh([self.HOST_BIN_PATH, 'stop', self._name])

    def state(self):
        ret = self._vm.ssh(
            [
                self.BIN_PATH, 'exec vdsmc systemctl status --lines=0',
                self._name
            ]
        )
        if ret.code == 0:
            return ServiceState.ACTIVE

        lines = [l.strip() for l in ret.out.split('\n')]
        loaded = [l for l in lines if l.startswith('Loaded:')].pop()

        if loaded.split()[1] == 'loaded':
            return ServiceState.INACTIVE

        ret = self._vm.ssh([self.HOST_BIN_PATH, 'status', self._name])
        if ret.code == 0:
            return ServiceState.ACTIVE

        lines = [l.strip() for l in ret.out.split('\n')]
        loaded = [l for l in lines if l.startswith('Loaded:')].pop()

        if loaded.split()[1] == 'loaded':
            return ServiceState.INACTIVE

        return ServiceState.MISSING



#
# Copyright 2016-2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#

from future.builtins import super
import functools
import logging
import os
import pwd
import time
import sys

import libvirt
from lxml import etree as ET
from textwrap import dedent

from lago import export, log_utils, sysprep, utils
from lago.utils import LagoException
from lago.config import config
from lago.plugins import vm as vm_plugin
from lago.plugins.vm import ExtractPathError
from lago.providers.libvirt import utils as libvirt_utils
from lago.providers.libvirt import cpu
from lago.validation import check_import

LOGGER = logging.getLogger(__name__)

if check_import('guestfs'):
    from lago import guestfs_tools
else:
    LOGGER.debug('guestfs not available')

LogTask = functools.partial(log_utils.LogTask, logger=LOGGER)
log_task = functools.partial(log_utils.log_task, logger=LOGGER)


class LocalLibvirtVMProvider(vm_plugin.VMProviderPlugin):
    def __init__(self, vm):
        super().__init__(vm)
        self._has_guestfs = 'lago.guestfs_tools' in sys.modules
        libvirt_url = config.get('libvirt_url')
        self.libvirt_con = libvirt_utils.get_libvirt_connection(
            name=self.vm.virt_env.uuid + libvirt_url,
            libvirt_url=libvirt_url,
        )
        self._libvirt_ver = self.libvirt_con.getLibVersion()

        caps_raw_xml = self.libvirt_con.getCapabilities()
        self._caps = ET.fromstring(caps_raw_xml)

        host_cpu = self._caps.xpath('host/cpu')[0]
        self._cpu = cpu.CPU(spec=self.vm._spec, host_cpu=host_cpu)

        # TO-DO: found a nicer way to expose these attributes to the VM
        self.vm.cpu_model = self.cpu_model
        self.vm.cpu_vendor = self.cpu_vendor

    def __del__(self):
        if self.libvirt_con is not None:
            self.libvirt_con.close()

    def start(self):
        super().start()
        if not self.defined():
            # the wait_suspend method is a work around for:
            # https://bugzilla.redhat.com/show_bug.cgi?id=1411025
            # 'LAGO__START_WAIT__SUSPEND' should be set to a float or integer
            # indicating how much time to sleep between the time the domain
            # is created in paused mode, until it is resumed.
            wait_suspend = os.environ.get('LAGO__START__WAIT_SUSPEND')
            dom_xml = self._libvirt_xml()
            LOGGER.debug('libvirt XML: %s\n', dom_xml)
            with LogTask('Starting VM %s' % self.vm.name()):
                if wait_suspend is None:
                    self._createXML(dom_xml)
                else:
                    LOGGER.debug('starting domain in paused mode')
                    try:
                        wait_suspend = float(wait_suspend)
                    except:
                        raise ValueError(
                            'LAGO__START__WAIT_SUSPEND value is not a number'
                        )
                    dom = self._createXML(
                        dom_xml, flags=libvirt.VIR_DOMAIN_START_PAUSED
                    )
                    time.sleep(wait_suspend)
                    dom.resume()
                    if not dom.isActive():
                        raise RuntimeError(
                            'failed to resume %s domain' % dom.name()
                        )

    def _createXML(self, dom_xml, flags=0):
        try:
            dom = self.libvirt_con.createXML(dom_xml, flags)
        except libvirt.libvirtError as e:
            # 38 for general system call failure
            # 18 for failure in libvirt's storage driver
            if e.get_error_code() == 38 and \
                e.get_error_domain() == 18 and \
                    'Cannot access storage file' in e.get_error_message():
                raise LagoLocalLibvirtVMProviderException(
                    dedent(
                        """
                        Lago couldn't create VM: '{vm_name}' because of
                        missing permissions.

                        This error usually occurs when 'qemu' user doesn't
                        have permissions to access the storage files of
                        the vm located at {images_path}

                        A solution for the problem will be to set the
                        'execute/search' bit for any directory in the following
                        hierarchy {images_path}.
                        This can be achieved with:

                            chmod o+x /path/to/directory


                        For more information please refer to Lago's
                        installation docs.

                        Original error message from libvirt:

                            {libvirt_msg}
                        """
                    ).format(
                        vm_name=self.vm.name(),
                        images_path=self.vm.virt_env.prefix.paths.images(),
                        libvirt_msg=e.get_error_message()
                    )
                )
            else:
                raise

        if not dom:
            raise RuntimeError('Failed to create Domain: {}'.format(dom_xml))

        return dom

    def stop(self):
        super().stop()
        if self.defined():
            self.vm._ssh_client = None
            with LogTask('Destroying VM %s' % self.vm.name()):
                self.libvirt_con.lookupByName(self._libvirt_name(), ).destroy()

    def shutdown(self, *args, **kwargs):
        super().shutdown(*args, **kwargs)

        self._shutdown(
            libvirt_cmd=libvirt.virDomain.shutdown,
            ssh_cmd=['poweroff'],
            msg='Shutdown'
        )

        try:
            with utils.ExceptionTimer(timeout=60 * 5):
                while self.defined():
                    time.sleep(1)
        except utils.TimerException:
            raise utils.LagoUserException(
                'Failed to shutdown vm: {}'.format(self.vm.name())
            )

    def reboot(self, *args, **kwargs):
        super().reboot(*args, **kwargs)

        self._shutdown(
            libvirt_cmd=libvirt.virDomain.reboot,
            ssh_cmd=['reboot'],
            msg='Reboot'
        )

    def _shutdown(self, libvirt_cmd, ssh_cmd, msg):
        """
        Choose the invoking method (using libvirt or ssh)
        to shutdown / poweroff the domain.

        If acpi is defined in the domain use libvirt, otherwise use ssh.

        Args:
            libvirt_cmd (function): Libvirt function the invoke
            ssh_cmd (list of str): Shell command to invoke on the domain
            msg (str): Name of the command that should be inserted to the log
                message.

        Returns
            None

        Raises:
            RuntimeError: If acpi is not configured an ssh isn't available
        """
        if not self.defined():
            return

        with LogTask('{} VM {}'.format(msg, self.vm.name())):
            dom = self.libvirt_con.lookupByName(self._libvirt_name())
            dom_xml = dom.XMLDesc()

            idx = dom_xml.find('<acpi/>')
            if idx == -1:
                LOGGER.debug(
                    'acpi is not enabled on the host, '
                    '{} using ssh'.format(msg)
                )
                # TODO: change the ssh timeout exception from runtime exception
                # TODO: to custom exception and catch it.
                self.vm.ssh(ssh_cmd)
            else:
                LOGGER.debug('{} using libvirt'.format(msg))
                libvirt_cmd(dom)

    def defined(self):
        dom_names = [dom.name() for dom in self.libvirt_con.listAllDomains()]
        return self._libvirt_name() in dom_names

    def bootstrap(self):
        with LogTask('Bootstrapping %s' % self.vm.name()):
            if self.vm._spec['disks'][0]['type'] != 'empty' and self.vm._spec[
                'disks'
            ][0]['format'] != 'iso':
                root_disk = self.vm._spec['disks'][0]['path']
                mappings = {
                    'eth{0}'.format(idx): utils.ipv4_to_mac(nic['ip'])
                    for idx, nic in enumerate(self.vm.spec['nics'])
                }
                public_ssh_key = self.vm.virt_env.prefix.paths.ssh_id_rsa_pub()

                sysprep.sysprep(
                    disk=root_disk,
                    mappings=mappings,
                    distro=self.vm.distro(),
                    root_password=self.vm.root_password(),
                    public_key=public_ssh_key,
                    iscsi_name=self.vm.iscsi_name(),
                    hostname=self.vm.name(),
                )

    def state(self):
        """
        Return a small description of the current status of the domain

        Returns:
            str: small description of the domain status, 'down' if it's not
            defined at all.
        """
        if not self.defined():
            return 'down'

        state = self.libvirt_con.lookupByName(self._libvirt_name()).state()
        return libvirt_utils.Domain.resolve_state(state)

    def create_snapshot(self, name):
        if self.vm.alive():
            self._create_live_snapshot(name)
        else:
            self._create_dead_snapshot(name)

        self.vm.save()

    def revert_snapshot(self, name):
        try:
            snap_info = self.vm._spec['snapshots'][name]
        except KeyError:
            raise RuntimeError(
                'No snapshot %s for %s' % (name, self.vm.name())
            )

        with LogTask('Reverting %s to snapshot %s' % (self.vm.name(), name)):

            was_defined = self.defined()
            if was_defined:
                self.stop()
            for disk, disk_template in zip(self.vm._spec['disks'], snap_info):
                os.unlink(os.path.expandvars(disk['path']))
                ret, _, _ = utils.run_command(
                    [
                        'qemu-img',
                        'create',
                        '-f',
                        'qcow2',
                        '-b',
                        disk_template['path'],
                        disk['path'],
                    ],
                    cwd=os.path.dirname(os.path.expandvars(disk['path'])),
                )
                if ret != 0:
                    raise RuntimeError('Failed to revert disk')

            self._reclaim_disks()
            if was_defined:
                self.start()

    def extract_paths(self, paths, ignore_nopath):
        """
        Extract the given paths from the domain

        Attempt to extract all files defined in ``paths`` with the method
        defined in :func:`~lago.plugins.vm.VMProviderPlugin.extract_paths`,
        if it fails, and `guestfs` is available it will try extracting the
        files with guestfs.

        Args:
            paths(list of tuples): files to extract in
                `[(src1, dst1), (src2, dst2)...]` format.
            ignore_nopath(boolean): if True will ignore none existing paths.

        Returns:
            None

        Raises:
            :exc:`~lago.plugins.vm.ExtractPathNoPathError`: if a none existing
                path was found on the VM, and `ignore_nopath` is False.
            :exc:`~lago.plugins.vm.ExtractPathError`: on all other failures.
        """

        try:

            super().extract_paths(
                paths=paths,
                ignore_nopath=ignore_nopath,
            )
        except ExtractPathError as err:
            LOGGER.debug(
                '%s: failed extracting files: %s', self.vm.name(), err.message
            )
            if self._has_guestfs:
                self.extract_paths_dead(paths, ignore_nopath)
            else:
                raise

    def extract_paths_dead(self, paths, ignore_nopath):
        """
        Extract the given paths from the domain using guestfs.
        Using guestfs can have side-effects and should be used as a second
        option, mainly when SSH is not available.

        Args:
            paths(list of str): paths to extract
            ignore_nopath(boolean): if True will ignore none existing paths.

        Returns:
            None

        Raises:
            :exc:`~lago.utils.LagoException`: if :mod:`guestfs` is not
                importable.
            :exc:`~lago.plugins.vm.ExtractPathNoPathError`: if a none existing
                path was found on the VM, and `ignore_nopath` is True.
            :exc:`~lago.plugins.vm.ExtractPathError`: on failure extracting
                the files.
        """
        if not self._has_guestfs:
            raise LagoException(
                ('guestfs module not available, cannot '
                 )('extract files with libguestfs')
            )

        LOGGER.debug(
            '%s: attempting to extract files with libguestfs', self.vm.name()
        )
        guestfs_tools.extract_paths(
            disk_path=self.vm.spec['disks'][0]['path'],
            disk_root=self.vm.spec['disks'][0]['metadata'].get(
                'root-partition', 'root'
            ),
            paths=paths,
            ignore_nopath=ignore_nopath
        )

    def export_disks(
        self,
        standalone,
        dst_dir,
        compress,
        collect_only=False,
        with_threads=True,
        *args,
        **kwargs
    ):
        """
        Export all the disks of self.

        Args:
            standalone (bool): if true, merge the base images and the layered
                image into a new file (Supported only in qcow2 format)
            dst_dir (str): dir to place the exported disks
            compress(bool): if true, compress each disk.
            collect_only(bool): If true, return only a dict which maps between
                the name of the vm to the paths of the disks that will be
                exported (don't export anything).
            with_threads(bool): If True, export disks in parallel

        Returns:
            (dict): which maps between the name of the vm to the paths of
            the disks that will be exported
        """
        vm_export_mgr = export.VMExportManager(
            disks=self.vm.disks,
            dst=dst_dir,
            compress=compress,
            with_threads=with_threads,
            standalone=standalone,
            *args,
            **kwargs
        )

        if collect_only:
            return {self.vm.name(): vm_export_mgr.collect_paths()}
        else:
            return {self.vm.name(): vm_export_mgr.export()}

    @vm_plugin.check_defined
    def interactive_console(self):
        """
        Opens an interactive console

        Returns:
            lago.utils.CommandStatus: result of the virsh command execution
        """
        virsh_command = [
            "virsh",
            "-c",
            config.get('libvirt_url'),
            "console",
            self._libvirt_name(),
        ]
        return utils.run_interactive_command(command=virsh_command, )

    @property
    def cpu_model(self):
        """
        VM CPU model

        Returns:
            str: CPU model

        """
        return self._cpu.model

    @property
    def cpu_vendor(self):
        """
        VM CPU Vendor

        Returns:
            str: CPU vendor
        """
        return self._cpu.vendor

    def _libvirt_name(self):
        return self.vm.virt_env.prefixed_name(self.vm.name())

    def _get_qemu_kvm_path(self):
        qemu_kvm_path = self._caps.findtext(
            "guest[os_type='hvm']/arch[@name='x86_64']/domain[@type='kvm']"
            "/emulator"
        )

        if not qemu_kvm_path:
            LOGGER.warning("hardware acceleration not available")
            qemu_kvm_path = self._caps.findtext(
                "guest[os_type='hvm']/arch[@name='x86_64']"
                "/domain[@type='qemu']/../emulator"
            )

        if not qemu_kvm_path:
            raise utils.LagoException('kvm executable not found')

        return qemu_kvm_path

    def _load_xml(self):

        args = {
            'distro': self.vm.distro(),
            'libvirt_ver': self._libvirt_ver,
            'name': self._libvirt_name(),
            'mem_size': self.vm.spec.get('memory', 16 * 1024),
            'qemu_kvm': self._get_qemu_kvm_path()
        }

        dom_raw_xml = libvirt_utils.get_domain_template(**args)

        parser = ET.XMLParser(remove_blank_text=True)
        return ET.fromstring(dom_raw_xml, parser)

    def _libvirt_xml(self):

        dom_xml = self._load_xml()

        for child in self._cpu:
            dom_xml.append(child)

        devices = dom_xml.xpath('/domain/devices')[0]

        disk = devices.xpath('disk')[0]
        devices.remove(disk)

        scsi_con_exists = False
        for disk_order, dev_spec in enumerate(self.vm._spec['disks']):

            # we have to make some adjustments
            # we use iso to indicate cdrom
            # but libvirt wants it named raw
            # and we need to use cdrom device
            disk_device = 'disk'
            bus = 'virtio'
            if dev_spec['format'] == 'iso':
                disk_device = 'cdrom'
                dev_spec['format'] = 'raw'
                bus = 'ide'
            # names converted

            # support virtio-scsi - sdX devices
            if dev_spec['dev'].startswith('sd'):
                bus = 'scsi'
                if not scsi_con_exists:
                    controller = ET.Element(
                        'controller',
                        type='scsi',
                        index='0',
                        model='virtio-scsi',
                    )
                    driver = ET.Element(
                        'driver',
                        queues='{}'.format(self.vm._spec.get('vcpu', 2)),
                    )
                    controller.append(driver)
                    devices.append(controller)
                    scsi_con_exists = True

            disk = ET.Element(
                'disk',
                type='file',
                device=disk_device,
            )

            disk.append(
                ET.Element(
                    'driver',
                    name='qemu',
                    type=dev_spec['format'],
                    discard='unmap',
                ),
            )

            serial = ET.SubElement(disk, 'serial')
            serial.text = "{}".format(disk_order + 1)
            disk.append(serial)

            disk.append(
                ET.Element(
                    'source',
                    file=os.path.expandvars(dev_spec['path']),
                ),
            )
            disk.append(
                ET.Element(
                    'target',
                    dev=dev_spec['dev'],
                    bus=bus,
                ),
            )
            devices.append(disk)

        for dev_spec in self.vm._spec['nics']:
            interface = ET.Element(
                'interface',
                type='network',
            )
            interface.append(
                ET.Element(
                    'source',
                    network=self.vm.virt_env.prefixed_name(
                        dev_spec['net'], max_length=15
                    ),
                ),
            )
            interface.append(
                ET.Element(
                    'model',
                    type='virtio',
                ),
            )
            if self._libvirt_ver > 3001001:
                mtu = dev_spec.get('mtu', '1500')
                if mtu != '1500':
                    interface.append(ET.Element(
                        'mtu',
                        size=str(mtu),
                    ))
            if 'ip' in dev_spec:
                interface.append(
                    ET.Element(
                        'mac', address=utils.ipv4_to_mac(dev_spec['ip'])
                    ),
                )
            devices.append(interface)

        return ET.tostring(dom_xml, pretty_print=True)

    def _create_dead_snapshot(self, name):
        raise RuntimeError('Dead snapshots are not implemented yet')

    def _create_live_snapshot(self, name):
        with LogTask(
            'Creating live snapshot named %s for %s' % (name, self.vm.name()),
            level='debug',
        ):

            self.vm.wait_for_ssh()
            self.vm.guest_agent().start()
            self.vm.ssh('sync'.split(' '))

            dom = self.libvirt_con.lookupByName(self._libvirt_name())
            dom_xml = ET.fromstring(dom.XMLDesc())
            disks = dom_xml.xpath('devices/disk')

            snapshot_xml = ET.fromstring(
                libvirt_utils.get_template('snapshot_template.xml')
            )
            snapshot_disks = snapshot_xml.xpath('disks')[0]

            for disk in disks:
                target_dev = disk.xpath('target')[0].attrib['dev']
                snapshot_disks.append(ET.Element('disk', name=target_dev))

            try:
                dom.snapshotCreateXML(
                    ET.tostring(snapshot_xml),
                    libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_DISK_ONLY
                    | libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_QUIESCE,
                )
            except libvirt.libvirtError:
                LOGGER.exception(
                    'Failed to create snapshot %s for %s',
                    name,
                    self.vm.name(),
                )
                raise

            snap_info = []
            new_disks = ET.fromstring(dom.XMLDesc()).xpath('devices/disk')
            for disk, xml_node in zip(self.vm._spec['disks'], new_disks):
                disk['path'] = xml_node.xpath('source')[0].attrib['file']
                disk['format'] = 'qcow2'
                snap_disk = disk.copy()
                snap_disk['path'] = xml_node.xpath(
                    'backingStore',
                )[0].xpath(
                    'source',
                )[0].attrib['file']
                snap_info.append(snap_disk)

            self._reclaim_disks()
            self.vm._spec['snapshots'][name] = snap_info

    def _reclaim_disk(self, path):
        qemu_uid = None
        try:
            qemu_uid = pwd.getpwnam('qemu').pw_uid
        except KeyError:
            pass
        if qemu_uid is not None and os.stat(path).st_uid == qemu_uid:
            utils.run_command(['sudo', '-u', 'qemu', 'chmod', 'a+rw', path])
        else:
            os.chmod(path, 0o0666)

    def _reclaim_disks(self):
        for disk in self.vm._spec['disks']:
            self._reclaim_disk(disk['path'])


class LagoLocalLibvirtVMProviderException(utils.LagoException):
    pass
